import time

from redis_lock.base import BaseSyncLock
from redis_lock.exceptions import LockNotOwnedError


class RedisSpinLock(BaseSyncLock):

    # keys: (Lock.name,)
    # args (Lock.token,)
    LUA_RELEASE = """
        if redis.call("GET", KEYS[1]) ~= ARGV[1] then
            return 0
        end
        redis.call("DEL", KEYS[1])
        return 1
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lua_release = self._client.register_script(self.LUA_RELEASE)

    def acquire(self, blocking: bool = True, sleep_time: float = 0.1) -> bool:
        """Try to acquire a lock

        Args:
            blocking: If `True`, wait for the lock to be acquired. If not,
                return the results immediately.
            sleep_time: The sleep time in seconds to wait for the next
                iteration if the lock is not accessible.

        Returns:
           bool: `True` if the lock was acquired, `False` otherwise.
        """
        timeout = self._blocking_timeout

        while True:
            current = time.time()
            if self._client.set(self.name, self.token, nx=True, ex=self._ex):
                return True
            elif not blocking:
                return False
            time.sleep(sleep_time)
            timeout -= (time.time() - current)
            if timeout < 0:
                return False

    def release(self) -> bool:
        """Release the owned lock

        Returns:
            bool: `True` if the lock was successfully released,
                `False` otherwise.
        """
        if not self.lua_release(keys=(self.name,), args=(self.token,)):
            raise LockNotOwnedError("Unable to release non-owned lock.")
        return True



import time

from redis import Redis
from redis.client import PubSub

from redis_lock.base import BaseSyncLock
from redis_lock.exceptions import LockNotOwnedError
from redis_lock.types import LockKey


class RedisLock(BaseSyncLock):

    unlock_message = b"ok"

    # keys: (Lock.name, Lock.channel_name)
    # args (Lock.token, Lock.unlock_message)
    LUA_RELEASE = """
        if redis.call("GET", KEYS[1]) ~= ARGV[1] then
            return 0
        end
        redis.call("DEL", KEYS[1])
        redis.call("PUBLISH", KEYS[2], ARGV[2]);
        return 1
    """

    def __init__(self, client: Redis, name: LockKey, *args, **kwargs):
        super().__init__(client, name, *args, **kwargs)
        self.lua_release = self._client.register_script(self.LUA_RELEASE)

    @property
    def channel_name(self) -> str:
        return f"rlock-channel:{self.name}"

    def _try_acquire(self) -> bool:
        if self._client.set(self.name, self.token, nx=True, ex=self._ex):
            return True
        return False

    def _subscribe_channel(self, pubsub: PubSub):
        if not pubsub.subscribed:
            pubsub.subscribe(self.channel_name)

    def _wait_for_message(self, pubsub: PubSub, timeout: int) -> bool:
        break_time = time.time() + timeout
        while True:
            message = pubsub.get_message(
                ignore_subscribe_messages=True, timeout=timeout
            )
            if not message and break_time < time.time():
                return False
            elif (
                message
                and message["type"] == "message"
                and message["data"] == self.unlock_message
            ):
                return True

    def acquire(self) -> bool:
        """Try to acquire a lock

        Returns:
            bool: `True` if the lock was acquired, `False` otherwise.
        """
        timeout = self._blocking_timeout
        if self._try_acquire():
            return True

        with self._client.pubsub() as pubsub:
            self._subscribe_channel(pubsub)
            stop_trying_at = time.time() + timeout
            while True:
                self._wait_for_message(pubsub, timeout=timeout)
                if stop_trying_at < time.time():
                    return False
                elif self._try_acquire():
                    return True

    def release(self) -> bool:
        """Release the owned lock

        Returns:
            bool: `True` if the lock was successfully released,
                `False` otherwise.
        """
        if not self.lua_release(
            keys=(self.name, self.channel_name),
            args=(self.token, self.unlock_message),
        ):
            raise LockNotOwnedError("Unable to release non-owned lock.")
        return True



from abc import ABC, abstractmethod
import uuid
from typing import Optional

from redis_lock.exceptions import AcquireFailedError, InvalidArgsError
from redis_lock.types import LockKey, RedisClient, TimeOutType


class BaseLock(ABC):

    default_blocking_timeout = 60

    def __init__(
        self,
        client: RedisClient,
        name: LockKey = None,
        blocking_timeout: int = default_blocking_timeout,
        expire_timeout: Optional[TimeOutType] = None,
    ):
        """Base Redis lock interface

        Args:
            client: A Redis client object.
            name: The name to be used as Redis key.
            blocking_timeout: The maximum blocking time in seconds to wait for
                a lock to be acquired.
            expire_timeout: The maximum lifetime of acquired lock.
        """
        self._client = client
        if not name:
            raise InvalidArgsError("A `name` argument should be required.")
        self._name = name
        self._uuid = uuid.uuid4().hex.encode()
        self._validate_timeout(blocking_timeout)
        self._blocking_timeout = blocking_timeout
        self._ex = expire_timeout

    @property
    def name(self) -> LockKey:
        return self._name

    @property
    def token(self) -> bytes:
        return self._uuid

    @staticmethod
    def _validate_timeout(blocking_timeout: int):
        if not blocking_timeout:
            raise InvalidArgsError(
                "A `blocking_timeout` argument should be provided at the "
                "time of initializing the `Lock` class."
            )
        elif (
            not isinstance(blocking_timeout, int)
            and not isinstance(blocking_timeout, float)
            or blocking_timeout < 0
        ):
            raise InvalidArgsError(
                "A `blocking_timeout` argument should be integer of float and "
                "cannot be negative."
            )


class BaseSyncLock(BaseLock):
    """Base Redis lock implementation"""

    def __enter__(self):
        if self.acquire():
            return self
        raise AcquireFailedError("Failed to acquire a lock.")

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    @abstractmethod
    def acquire(self, *args, **kwargs):
        """Try to acquire a lock"""
        raise NotImplementedError(
            "The `acquire` method should be implemented!"
        )

    @abstractmethod
    def release(self):
        """Release the owned lock"""
        raise NotImplementedError(
            "The `release` method should be implemented!"
        )



import re
from setuptools import find_packages, setup
from typing import List


def get_version() -> str:
    with open("redis_lock/__init__.py", "r") as f:
        pattern = r"__version__ = \"([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})\""
        version = re.findall(pattern, f.read())
        return version[0]


def get_requirements(filename: str) -> List:
    with open(filename, "r") as f:
        lines = f.readlines()
    return [line.strip() for line in lines]


setup(
    name="redis-lock-py",
    version=get_version(),
    license="MIT",
    author="Minjae Park",
    author_email="miintto.log@google.com",
    description="Redis distributed lock implementation for Python "
                "base on Pub/Sub messaging",
    long_description=open("README.md").read().strip(),
    long_description_content_type="text/markdown",
    url="https://github.com/miintto/redis-lock-py",

    packages=find_packages(include=["redis_lock"]),

    python_requires=">=3.7",
    install_requires=[
        "redis>=4.2.0"
    ],

    keywords=[
        "redis", "distributed-lock", "python", "pubsub-messages", "asyncio"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    tests_require=["pytest", "pytest-asyncio"],
)
